@@ -2,7 +2,7 @@
 *     xfpt - Simple ASCII->Docbook processor     *
 *************************************************/
 
-/* Copyright (c) University of Cambridge, 2012 */
+/* Copyright (c) University of Cambridge, 2020 */
 /* Written by Philip Hazel. */
 
 /* Error handling routines */
@@ -75,10 +75,15 @@ static error_struct error_data[] = {
 { ec_serious,  "ampersand found at end of line or string - ignored" },
 { ec_serious,  "\"begin\" or \"end\" expected, but \"%s\" found" },
 { ec_serious,  "\".nest begin\" too deeply nested" },
-{ ec_serious,  "\".nest end\" incorrectly nested" }
+{ ec_serious,  "\".nest end\" incorrectly nested" },
+{ ec_serious,  "missing delimiter after $=%d at end of line" },
+/* 30-34 */
+{ ec_serious,  "bad macro argument substitution: %s follows \"%s\"" },
+{ ec_serious,  "binary zero in input ignored" },
+{ ec_disaster, "input sources too deeply nested" }
 };
 
-#define error_maxerror 28
+#define error_maxerror (int)(sizeof(error_data)/sizeof(error_struct))
 
 
 
@@ -177,6 +182,7 @@ else if (ec > ec_warning)
 if (ec >= ec_failed)
   {
   (void)fprintf(stderr, "** xfpt abandoned\n");
+  suppress_warnings = TRUE;  /* Avoid unclosed warnings */
   exit(EXIT_FAILURE);
   }
 
