@@ -2,7 +2,7 @@
 *     xfpt - Simple ASCII->Docbook processor     *
 *************************************************/
 
-/* Copyright (c) University of Cambridge, 2018 */
+/* Copyright (c) University of Cambridge, 2020 */
 /* Written by Philip Hazel. */
 
 /* This module contains code for reading the input. */
@@ -19,7 +19,9 @@
 /* This is the place where macro arguments are substituted. In a section
 delimited by .eacharg/.endeach, the variable macro_argbase is set to the first
 of the relative arguments. This function is also called from para.c in order to
-handle inline macro calls.
+handle inline macro calls. There will always be a zero at the end of the input
+line, but we must allow for zeroes inside the line, so this cannot be used to
+detect end of line.
 
 Arguments:
   p         the macro input line
@@ -53,6 +55,10 @@ while (*p != 0)
 
   if (*p != '$') { *b++ = *p++; continue; }
 
+  /* If dollar is at the end of the string, treat as literal. */
+
+  if (p[1] == 0) { *b++ = '$'; break; }
+
   /* If the character after $ is another $, insert a literal $. */
 
   if (p[1] == '$') { p++; *b++ = *p++; continue; }
@@ -67,7 +73,7 @@ while (*p != 0)
     if (macro_argbase == NULL)       /* Not in a .eacharg section */
       {
       error(18);
-      *b++ = '+';
+      *b++ = '$';
       *b++ = *p++;
       continue;
       }
@@ -82,15 +88,23 @@ while (*p != 0)
     p++;
     if (!isdigit(p[1]))
       {
-      error(17, p[1], "$=");
+      if (p[1] == 0 || p[1] == '\n') error(30, "end of line", "$=");
+        else error(17, p[1], "$=");
       *b++ = '$';
       *b++ = *p++;
       continue;
       }
     while (isdigit(*(++p))) argn = argn * 10 + *p - '0';
 
-    optend = *p++;
+    if (*p == 0 || *p == '\n')
+      {
+      error(29, argn);
+      break;
+      }
 
+    /* Pick up the delimiter and check for the argument. */
+
+    optend = *p++;
     arg = argbase;
     for (i = 1; i < argn; i++)
       {
@@ -101,7 +115,7 @@ while (*p != 0)
     if (arg == NULL || arg->string[0] == 0)
       {
       while (*p != 0 && *p != optend) p++;
-      if (*p == optend) p++;
+      if (*p != 0) p++;
       }
 
     continue;
@@ -111,11 +125,11 @@ while (*p != 0)
 
   if (!isdigit(p[1]))
     {
-    error(17, p[1], "$");
+    if (*p == 0 || *p == '\n') error(30, "end of line", "$");
+      else error(17, p[1], "$");
     *b++ = *p++;
     continue;
     }
-
   while (isdigit(*(++p))) argn = argn * 10 + *p - '0';
 
   /* Handle $0 - currently no meaning */
@@ -157,6 +171,47 @@ while (*p != 0)
 
 
 
+/*************************************************
+*          Read a line from a file               *
+*************************************************/
+
+/* Input is read by fgetc() rather than fgets() so that we can detect and
+complain about binary zeroes, which have been thrown at xfpt by fuzzers.
+
+Arguments:
+  f           the input file
+  buffer      where to read into
+  blen        length of buffer
+
+Returns:      number of chars read or -1 at EOF
+*/
+
+static int
+get_nextline(FILE *f, uschar *buffer, int blen)
+{
+uschar *p = buffer;
+uschar *pend = buffer + blen;
+
+for (;;)
+  {
+  int c;
+  if (p >= pend - 1) break;
+  c = fgetc(f);
+  if (c == EOF) return -1;
+  if (c == 0)
+    {
+    error(31);
+    continue;
+    }
+  *p++ = c;
+  if (c == '\n') break;
+  }
+*p = 0;
+return (int)(p - buffer);
+}
+
+
+
 /*************************************************
 *         Get the next line of input             *
 *************************************************/
@@ -252,7 +307,8 @@ for (;;)
 
   else
     {
-    if (Ufgets(inbuffer, INBUFFSIZE, istack->file) == NULL)
+    len = get_nextline(istack->file, inbuffer, INBUFFSIZE);
+    if (len < 0)
       {
       istackstr *prev = istack->prev;
       fclose(istack->file);
@@ -262,9 +318,7 @@ for (;;)
     else
       {
       istack->linenumber++;
-
       q = inbuffer;
-      len = Ustrlen(q);
 
       for (;;)
         {
@@ -276,14 +330,13 @@ for (;;)
         q = p - 3;
         *q = 0;
 
-        if (istack == NULL ||
-            Ufgets(q, INBUFFSIZE - (q - inbuffer), istack->file) == NULL)
-          break;
+        if (istack == NULL) break;
+        len = get_nextline(istack->file, q, INBUFFSIZE - (q - inbuffer));
+        if (len < 0) break;
 
         istack->linenumber++;
         p = q;
-        while (*p == ' ' || *p == '\t') p++;
-        len = Ustrlen(p);
+        while (*p == ' ' || *p == '\t') { p++; len--; }
         if (p > q) memmove(q, p, len + 1);
         }
 
