@@ -2,7 +2,7 @@
 *     xfpt - Simple ASCII->Docbook processor     *
 *************************************************/
 
-/* Copyright (c) University of Cambridge, 2018 */
+/* Copyright (c) University of Cambridge, 2020 */
 /* Written by Philip Hazel. */
 
 /* This module contains the main program and initialization functions. */
@@ -17,6 +17,7 @@
 
 static uschar *xfpt_filename = NULL;
 static uschar *out_filename = NULL;
+static BOOL out_filename_onheap = FALSE;
 
 
 
@@ -100,6 +101,7 @@ if (xfpt_filename != NULL && out_filename == NULL)
   uschar *p;
   int len = Ustrlen(xfpt_filename);
   out_filename = misc_malloc(len + 5);
+  out_filename_onheap = TRUE;
   Ustrcpy(out_filename, xfpt_filename);
   if ((p = Ustrrchr(out_filename, '.')) != NULL) len = p - out_filename;
   Ustrcpy(out_filename + len, ".xml");
@@ -110,6 +112,140 @@ return TRUE;
 
 
 
+/*************************************************
+*          Free a tree node and its children     *
+*************************************************/
+
+static void
+free_tree(tree_node *t)
+{
+if (t == NULL) return;
+free_tree(t->left);
+free_tree(t->right);
+free(t->data);
+free(t);
+}
+
+
+
+/*************************************************
+*          Clean up function at exit             *
+*************************************************/
+
+static void
+clean_up(void)
+{
+int warnpop = 0;
+macrodef *md = macrolist;
+macroexe *me = macrocurrent;
+istackstr *ist = istack;
+flagstr *f = flaglist;
+
+/* Empty the stack of pushed texts. This must be done first because it calls
+para_process(), which needs access to flags, entities, etc. */
+
+while (pushed != NULL)
+  {
+  pushstr *temp;
+  if (!suppress_warnings)
+    {
+    if (pushed->check != 0)
+      {
+      if (warnpop++ == 0)
+        fprintf(stderr,
+           "** Warning: one or more items were left unclosed at the end of processing.\n"
+           "   The numbers are the lines in the original file %s from where\n"
+           "   the items were generated:\n",
+           ((xfpt_filename == NULL)? "(stdin)" : (char *)xfpt_filename));
+      if (pushed->macname == NULL)
+        fprintf(stderr, "%d: %s\n", pushed->check, pushed->string);
+      else
+        fprintf(stderr, "%d: .%s\n", pushed->check, pushed->macname);
+      if (warnpop > 10)
+        {
+        fprintf(stderr, "... too many to list\n");
+        suppress_warnings = TRUE;
+        }
+      }
+    }
+  para_process(pushed->string);
+  (void)fprintf(outfile, "\n");
+  temp = pushed;
+  pushed = pushed->next;
+  free(temp);
+  }
+
+/* Release macro definitions */
+
+while (md != NULL)
+  {
+  macrodef *mdnext = md->next;
+  argstr *as = md->args;
+
+  while (as != NULL)
+    {
+    argstr *next = as->next;
+    free(as->string);
+    free(as);
+    as = next;
+    }
+
+  as = md->lines;
+  while (as != NULL)
+    {
+    argstr *next = as->next;
+    free(as->string);
+    free(as);
+    as = next;
+    }
+
+  free(md->name);
+  free(md);
+  md = mdnext;
+  }
+
+/* Release macro execution blocks */
+
+while (me != NULL)
+  {
+  macroexe *menext = me->prev;
+  macro_free(me);
+  me = menext;
+  }
+
+/* Release input source blocks */
+
+while (ist != NULL)
+  {
+  istackstr *prev = ist->prev;
+  if (ist->file != NULL && ist->file != stdin) fclose(ist->file);
+  free(ist);
+  ist = prev;
+  }
+
+/* Release flag definitions */
+
+while (f != NULL)
+  {
+  flagstr *fnext = f->next;
+  flag_free(f);
+  f = fnext;
+  }
+
+/* Free entity definitions */
+
+free_tree(entities);
+
+/* Release general buffers, close output unless stdout, and free its name if
+necessary. */
+
+free(inbuffer);
+free(parabuffer);
+if (outfile != stdout) fclose(outfile);
+if (out_filename_onheap) free(out_filename);
+}
+
+
 /*************************************************
 *          Entry point and main program          *
 *************************************************/
@@ -118,10 +254,10 @@ int
 main(int argc, char **argv)
 {
 BOOL para_unfinished[MAXNEST+1];
-int warnpop = 0;
-flagstr *f, *fnext;
 uschar *p, *q;
 
+atexit(clean_up);
+
 if (!xfpt_decode_arg(argc, argv)) return EXIT_FAILURE;
 
 inbuffer = misc_malloc(INBUFFSIZE);
@@ -221,49 +357,8 @@ while ((p = read_nextline()) != NULL)
     }
   }
 
-/* Empty the stack of pushed texts, close the output, free memory, and we are
-done. */
-
-while (pushed != NULL)
-  {
-  pushstr *temp;
-  if (!suppress_warnings)
-    {
-    if (pushed->check != 0)
-      {
-      if (warnpop++ == 0)
-        fprintf(stderr,
-           "** Warning: one or more items were left unclosed at the end of processing.\n"
-           "   The numbers are the lines in the original file %s from where\n"
-           "   the items were generated:\n",
-           ((xfpt_filename == NULL)? "(stdin)" : (char *)xfpt_filename));
-      if (pushed->macname == NULL)
-        fprintf(stderr, "%d: %s\n", pushed->check, pushed->string);
-      else
-        fprintf(stderr, "%d: .%s\n", pushed->check, pushed->macname);
-
-      if (warnpop > 10)
-        {
-        fprintf(stderr, "... too many to list\n");
-        suppress_warnings = TRUE;
-        }
-      }
-    }
-  para_process(pushed->string);
-  (void)fprintf(outfile, "\n");
-  temp = pushed;
-  pushed = pushed->next;
-  free(temp);
-  }
-
-(void)fclose(outfile);
-
-for (f = flaglist; f != NULL; f = fnext)
-  {
-  fnext = f->next;
-  flag_free(f);
-  f = fnext;
-  }
+/* Clean up of heap memory and file closure happens in the clean_up() function
+so that it happens as well on hard errors. */
 
 return return_code;
 }
